'' =================================================================================================
''
''   File....... jm_serial_tx.spin
''   Purpose.... True mode serial transmitter with buffer
''   Author..... Jon "JonnyMac" McPhalen
''               Copyright (C) 2009-2022 Jon McPhalen
''               -- see below for terms of use
''   E-mail..... jon.mcphalen@gmail.com
''   Started....
''   Updated.... 16 DEC 2022
''
'' =================================================================================================


con { fixed io pins }

  RX1      = 31  { I }                                          ' serial / programming
  TX1      = 30  { O }

  SDA1     = 29  { I/O }                                        ' i2c / eeprom
  SCL1     = 28  { I/O }


con { buffer setting }

  BUF_SIZE = 64                                                 ' 16, 32, 64, 128, 256, or 512
  BUF_MASK = BUF_SIZE - 1


con { pst formatting }

' Control code for Parallax Serial Terminal

  HOME    =  1
  CRSR_XY =  2         ' next two bytes in stream are x and y positions values
  CRSR_LF =  3
  CRSR_RT =  4
  CRSR_UP =  5
  CRSR_DN =  6
  BELL    =  7
  BKSP    =  8
  TAB     =  9
  LF      = 10
  CLR_EOL = 11         ' clear to end of line
  CLR_DN  = 12         ' clear everything after cursor
  F_FEED  = 12
  CR      = 13
  CRSR_X  = 14         ' next byte in stream is x position value
  CRSR_Y  = 15         ' next byte in stream is y position value
  CLS     = 16


obj

  nstr : "jm_nstr"                                              ' formatting routines


var

  long  cog

  long  pin                                                     ' tx pin
  long  bitticks                                                ' system tick in one bit
  long  bufaddr                                                 ' hub address of buffer
  long  txhead                                                  ' head pointer (0..BUF_SIZE-1)
  long  txtail                                                  ' tail pointer (0..BUF_SIZE-1)

  byte  buffer[BUF_SIZE]                                        ' transmit buffer


pub null

'' This is not an application


pub start(txd, baud)

'' Creates true mode transmit UART on pin txd at baud rate

  stop                                                          ' stop cog if running

  pin      := txd                                               ' set parameters
  bitticks := clkfreq / baud
  bufaddr  := @buffer

  dira[pin] := 0                                                ' release in this cog

  cog := cognew(@txserial, @pin) + 1                            ' start tx cog

  return cog


pub stop

'' Stops serial TX driver; frees a cog

  if (cog)
    cogstop(cog - 1)
    cog := 0


pub tx(b)

'' Emit byte b
'' -- may wait for room in buffer

  repeat
  until (txtail <> (txhead + 1) & BUF_MASK)                     ' wait for room in buffer

  buffer[txhead] := b                                           ' move byte to buffer
  txhead := ++txhead & BUF_MASK                                 ' increment buffer index, wrap if needed


pub txn(b, n)

'' Emit byte n times

  repeat n
    tx(b)


pub txflush

'' Wait for transmit buffer to empty, then wait for final byte to transmit

  repeat
  until (txtail == txhead)

  repeat 11                                                     ' start + 8 + 2
    waitcnt(bitticks + cnt)                                     ' time enough for final byte


pub str(p_str)

'' Emit a string
'' -- p_str in pointer to 0-terminated string

  repeat strsize(p_str)                                         ' loop through string
    tx(byte[p_str++])                                           ' tx char, advance pointer


pub substr(p_str, len) | b

'' Emit len characters of string at p_str
'' -- aborts if end of string detected

  repeat len
    b := byte[p_str++]                                          ' get a byte from string
    if (b > 0)                                                  ' if not at end
      tx(b)                                                     '   send it
    else
      quit


pub fmt_number(value, radix, digits, width, pad)

'' Emit value converted to number in padded field

  str(nstr.fmt_number(value, radix, digits, width, pad))


pub fxdec(value, digits)

'' Emit value as decimal using fixed # of digits
'' -- digits = 0 for auto width
'' -- may add leading zeros

  str(nstr.dec(value, digits))


pub fxhex(value, digits)

'' Emit value as hexadecimal using fixed # of digits
'' -- digits = 0 for auto width
'' -- may add leading zeros

  str(nstr.itoa(value, 16, digits))


pub fxbin(value, digits)

'' Emit value as binary using fixed # of digits
'' -- digits = 0 for auto width
'' -- may add leading zeros

  str(nstr.itoa(value, 2, digits))


dat

                        org     0

txserial                mov     t1, par                         ' get address of parameters
                        rdlong  t2, t1                          ' read tx pin
                        mov     txmask, #1                      ' convert to mask
                        shl     txmask, t2
                        or      outa, txmask                    ' set to idle (mark state)
                        or      dira, txmask

                        add     t1, #4                          ' next parameter
                        rdlong  txbit1x0, t1                    ' read bit timing

                        add     t1, #4
                        rdlong  txbufpntr, t1                   ' get buffer address

                        add     t1, #4
                        mov     txheadpntr, t1                  ' save pointer to txhead

                        add     t1, #4
                        mov     txtailpntr, t1                  ' save pointer to txtail


checkbuf                rdlong  t1, txheadpntr                  ' r1 = txhead
                        rdlong  t2, txtailpntr                  ' r2 = txtail
                        cmp     t1, t2                  wz      ' equal? (empty buffer)
        if_e            jmp     #checkbuf                       ' if yes, keep waiting

gettail                 mov     t1, txbufpntr                   ' r1 := @buffer[0]
                        add     t1, t2                          ' r1 := @buffer[txtail]
                        rdbyte  txwork, t1                      ' txwork := txbuf[txtail]

updatetail              add     t2, #1                          ' inc txtail
                        and     t2, #BUF_MASK                   ' wrap to 0 if necessary
                        wrlong  t2, txtailpntr                  ' save


transmit                or      txwork, STOP_BITS               ' set stop bit(s)
                        shl     txwork, #1                      ' add start bit
                        mov     txtimer, txbit1x0               ' load bit timing
                        add     txtimer, cnt                    ' sync with system counter

                        { unrolled for best speed }

                        shr     txwork, #1              wc      ' move start bit to C
                        muxc    outa, txmask                    ' output the bit
                        waitcnt txtimer, txbit1x0               ' let timer expire, reload

                        shr     txwork, #1              wc      ' move bit0 to C
                        muxc    outa, txmask
                        waitcnt txtimer, txbit1x0

                        shr     txwork, #1              wc      ' move bit1 to C
                        muxc    outa, txmask
                        waitcnt txtimer, txbit1x0

                        shr     txwork, #1              wc      ' move bit2 to C
                        muxc    outa, txmask
                        waitcnt txtimer, txbit1x0

                        shr     txwork, #1              wc      ' move bit3 to C
                        muxc    outa, txmask
                        waitcnt txtimer, txbit1x0

                        shr     txwork, #1              wc      ' move bit4 to C
                        muxc    outa, txmask
                        waitcnt txtimer, txbit1x0

                        shr     txwork, #1              wc      ' move bit5 to C
                        muxc    outa, txmask
                        waitcnt txtimer, txbit1x0

                        shr     txwork, #1              wc      ' move bit6 to C
                        muxc    outa, txmask
                        waitcnt txtimer, txbit1x0

                        shr     txwork, #1              wc      ' move bit7 to C
                        muxc    outa, txmask
                        waitcnt txtimer, txbit1x0

                        shr     txwork, #1              wc      ' 1st stop bit
                        muxc    outa, txmask
                        waitcnt txtimer, txbit1x0

                        shr     txwork, #1              wc      ' 2nd stop bit
                        muxc    outa, txmask
                        waitcnt txtimer, txbit1x0

                        jmp     #checkbuf


' -------------------------------------------------------------------------------------------------

STOP_BITS               long    $FFFF_FF00

txmask                  res     1                               ' mask for tx pin
txbit1x0                res     1                               ' ticks per bit
txheadpntr              res     1                               ' pointer to head position
txtailpntr              res     1                               ' pointer to tail position
txbufpntr               res     1                               ' pointer to buffer[0]

txwork                  res     1                               ' rx byte input
txtimer                 res     1                               ' timer for bit sampling
txcount                 res     1

t1                      res     1
t2                      res     1

                        fit     496


con { license }

{{

  Terms of Use: MIT License

  Permission is hereby granted, free of charge, to any person obtaining a copy of this
  software and associated documentation files (the "Software"), to deal in the Software
  without restriction, including without limitation the rights to use, copy, modify,
  merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
  permit persons to whom the Software is furnished to do so, subject to the following
  conditions:

  The above copyright notice and this permission notice shall be included in all copies
  or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
  INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
  PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
  HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
  CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
  OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

}}